<?php class baseobject implements crudable {

	protected static $fields=array();
	protected static $helpTexts=array();
	
	public static function getFieldValidations(){
		return static::$fields; //static, not self - use the child class's.
	}
	public static function getFieldHelpTexts(){
		return static::$helpTexts; //static, not self - use the child class's.
	}
	
	public static $defaultSortOrder='name ASC';

	/**
	 * The terms used where this class is included in a "get by name like" search.
	 * By default, this is the name property. However, in some cases it may be useful
	 * to override this so that search is done one a more relevant property.
	 * @var string
	 */
	protected static $searchClause='[[classname]].name';
	
	protected static $adminSelect= /** @lang text */
        'SELECT SQL_CALC_FOUND_ROWS [[classname]].*,project.name AS projectname, project.isarchived AS isarchived FROM [[classname]], project WHERE project.id=[[classname]].projectid ';
	protected static $normalSelect= /** @lang text */
        'SELECT SQL_CALC_FOUND_ROWS [[classname]].*,project.name AS projectname, project.isarchived AS isarchived FROM [[classname]], project WHERE project.id=[[classname]].projectid ';

    /**
     * Return the SQL restricting records to those on which the user has the specified permission
     * @param string $accessType One of "read", "update", "delete"
     * @param bool $forceSharedProject
     * @return string
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
	protected static function getProjectClause($accessType, $forceSharedProject=false){
		return database::getProjectClause($accessType, $forceSharedProject);
	}

    /**
     * Creates a record in the baseobject table and returns its ID.
     * @param int $projectId The ID of the project for the new record
     * @param string $className
     * @return int The database ID of the baseobject row
     * @throws BadRequestException
     * @throws ForbiddenException if the user can read but not create records in the project
     * @throws NotFoundException if the project does not exist or the user cannot read it
     * @throws ServerException
     */
	protected static function createBaseObject($projectId, $className){
		//$className=get_called_class();
		if(empty($projectId)){
			throw new BadRequestException('Project ID not provided.');
		}
		if(!session::isAdmin()){
			$forceSharedProject=(isset($className::$sharedProjectOverride) && $className::$sharedProjectOverride);
			$projects=session::getProjectPermissions('create',$forceSharedProject);
			if(!in_array($projectId,$projects)){
				session::refreshProjectPermissions();
				$projects=session::getProjectPermissions('create',$forceSharedProject);
			}
			if(!in_array($projectId,$projects)){
			    $projects=session::getProjectPermissions('read',$forceSharedProject);
				if(!in_array($projectId,$projects)){
					throw new NotFoundException("Either the project does not exist or you do not have permission to see it");
				} else {
					throw new ForbiddenException("You do not have permission to create in that project");
				}
			}
		}
		$sqlStatement='INSERT INTO baseobject(projectid,objecttype,creator,createtime) VALUES(:projectid,:objecttype,:userid,:now)';
		$now=new DateTime();
		$now->setTimezone(new DateTimeZone('GMT'));
		$params=array(
			':projectid'=>$projectId,
			':objecttype'=>$className,
			':userid'=>session::getUserId(),
			':now'=>$now->format('Y-m-d H:i:s')
		);
		database::query($sqlStatement,$params);
		return database::getLastInsertId();
	}

    /**
     * Returns the record with the specified ID.
     * @param int $id The ID of the desired record
     * @return array The record
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function getById($id){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$sqlStatement.=' AND [[classname]].id=:id';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$params=array(':id'=>$id);
		return database::queryGetOne($sqlStatement,$params);
	}


    /**
     * Returns the record with the specified name.
     * @param string $name The name of the desired record
     * @return array The record
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function getByName($name){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$sqlStatement.=' AND [[classname]].name=:name';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$params=array(':name'=>$name);
		return database::queryGetOne($sqlStatement,$params);
	}

    /**
     * @param string $searchTerms
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function getByNameLike($searchTerms,$request=array()){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$searchTerms=strtolower($searchTerms);
		$sqlStatement=str_replace('SELECT ', 'SELECT "[[classname]]" AS objecttype, ', $sqlStatement);
		$sqlStatement.=' AND LOWER([[classname]].name) LIKE :terms ';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$sqlStatement.=' ORDER BY INSTR(LOWER([[classname]], :terms2) '; 
		$sqlStatement.=database::getLimitClause($request);
		$params=array(
				':terms'=>'%'.$searchTerms.'%',
				':terms2'=>'%'.$searchTerms.'%'
		);
		return database::queryGetAll($sqlStatement,$params);
	}

    /**
     * @param $searchTerms
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws ServerException
     */
	public static final function getByNameLikeMultiClass($searchTerms, $request=array()){
		if(!isset($request['classes'])){
			throw new BadRequestException('Specify classes to search for in querystring, e.g., ?classes=type1,type2,type3');
		}
		$originalSearchTerms=$searchTerms;
		$searchTerms=strtolower($searchTerms);
		//Replace UI wildcards with SQL ones
        $searchTerms=str_replace('?','_', $searchTerms);
        $searchTerms=str_replace('*','%', $searchTerms);

        //Create wildcarded inner search term, replacing potentially troublesome double wildcards
        $innerSearchTerms='%'.$searchTerms.'%';
        $innerSearchTerms=str_replace('%%','%', $innerSearchTerms);
        $innerSearchTerms=str_replace('%_','%', $innerSearchTerms);
        $innerSearchTerms=str_replace('_%','%', $innerSearchTerms);

		$classes=explode(',', $request['classes']);
		$parts=array();
		$params=array();
		$counter=1;
		foreach($classes as $className){
			if(!class_exists($className)){
				throw new BadRequestException('Class '.$className.' does not exist');
			}
			if(session::isAdmin()){
				$sqlStatement=static::$adminSelect;
			} else {
				$sqlStatement=static::$normalSelect;
			}

 			if(isset($className::$searchClause)){
 			    $searchClause=$className::$searchClause;
 			} else {
 			    $searchClause=baseobject::$searchClause;
 			}
			
			$sqlStatement=str_replace('SQL_CALC_FOUND_ROWS', '', $sqlStatement);
			$sqlStatement=str_replace('SELECT ', 'SELECT "[[classname]]" AS objecttype, ', $sqlStatement);
			//$sqlStatement=str_replace('[[classname]].*', '[[classname]].id, [[classname]].name, [[classname]].projectid', $sqlStatement);
			$sqlStatement=str_replace('[[classname]].*', '[[classname]].id, '.$searchClause.' AS name, [[classname]].projectid', $sqlStatement);
			//$sqlStatement.=' AND LOWER([[classname]].name) LIKE LOWER(:terms'.$counter.') ';
			$sqlStatement.=' AND LOWER('.$searchClause.') LIKE LOWER(:terms'.$counter.') ';
			$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
			$sqlStatement.=forward_static_call_array(array($className, 'getProjectClause'), array('read'));
			$parts[]=$sqlStatement;
			$params[':terms'.$counter]=$innerSearchTerms;
			$counter++;
		}

		$counter=1;
		$cases='';
		foreach ($classes as $c){
		    $cases.=' WHEN objecttype="'.$c.'" THEN '.$counter;
		    $counter++;
        }
        /** @noinspection SqlResolve */
        $sqlStatement='SELECT objecttype,id,name,projectid,projectname FROM ('.
		    implode(' UNION ALL ',$parts).
		    ') AS combinedresults ORDER BY CASE '.$cases.' END, INSTR(LOWER(name), :outerterm), objecttype, LOWER(name)';
		$sqlStatement.=database::getLimitClause($request);
		$params[':outerterm']=$searchTerms;
		$result=database::queryGetAll($sqlStatement,$params);
		$result['searchterms']=$originalSearchTerms;
		return $result;
	}

    /**
     * Generates the SQL condition for whether objects can be seen in a multi-class search. The default implementation
     * returns database::getProjectClause("read"), something line ' AND project.id IN(1,3,4) '. Any overriding implementation
     * must produce similar SQL for the search query to be valid.
     * @return string
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function getReadConditionsForMultiClassSearch(){
		return static::getProjectClause('read');
	}

    /**
     * Returns the records with the specified property.
     * @param string $key The name of the property
     * @param string $value The value of the property
     * @param array $request The request object, used for filtering, limit clause, etc.
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException if the record does not exist or the user cannot read it
     * @throws ServerException
     */
	public static function getByProperty($key,$value,$request=array()){
		$className=get_called_class();
		//verify key in fields array
		$keys=array_keys(self::getFieldValidations());
	
		if(!in_array($key, $keys) &&'projectid'!=$key){
			throw new BadRequestException('Property '.$key.' not recognised on '.$className);
		}
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$sqlStatement.=' AND [[classname]].'.$key.'<=>:val ';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		//		$sqlStatement='SELECT x.*,project.name AS projectname, project.isarchived AS isarchived FROM '.$className.' AS x, project WHERE project.id=x.projectid AND x.'.$key.'=:val';
		$sqlStatement.=static::getProjectClause('read');
        $sqlStatement.=database::getFilterClause($request, $className);
        $sqlStatement.=database::getOrderClause($request, $className);
		$sqlStatement.=database::getLimitClause($request);
		$params=array(':val'=>$value);
		return database::queryGetAll($sqlStatement,$params);
	}

    /**
     * Returns the records with the specified property.
     * @param array $keyValuePairs
     * @param array $request The request object, used for filtering, limit clause, etc.
     * @return array
     * @throws BadRequestException
     * @throws ServerException
     * @throws NotFoundException
     */
	public static function getByProperties($keyValuePairs,$request=array()){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$keys=array_keys(self::getFieldValidations());
		$params=array();
		foreach($keyValuePairs as $key=>$value){
			if(!in_array($key, $keys) &&'projectid'!=$key){
				throw new BadRequestException('Property '.$key.' not recognised on '.$className);
			}
			$sqlStatement.=' AND [[classname]].'.$key.'<=>:'.$key.' ';
			$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
			$params[':'.$key]=$value;
		}
		$sqlStatement.=static::getProjectClause('read');
        $sqlStatement.=database::getFilterClause($request, $className);
		$sqlStatement.=database::getOrderClause($request, $className);
		$sqlStatement.=database::getLimitClause($request);
		return database::queryGetAll($sqlStatement,$params);
	}

    /**
     * Returns all records, subject to any limits or filters in the request.
     * @param array $request The request object, used for filtering, limit clause, etc.
     * @return array
     * @throws BadRequestException
     * @throws ServerException
     * @throws NotFoundException
     */
	public static function getAll($request=array()){
		$className=get_called_class();
		//verify key in fields array
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		//$sqlStatement='SELECT x.*,project.name AS projectname, project.isarchived AS isarchived FROM '.$className.' AS x, project WHERE project.id=x.projectid';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
        $sqlStatement.=database::getFilterClause($request, $className);
		$sqlStatement.=database::getOrderClause($request, $className);
		$sqlStatement.=database::getLimitClause($request);
		$params=array();
		return database::queryGetAll($sqlStatement,$params);
	}

    /**
     * Gets the files associated with the record.
     * @param int $id The ID of the parent record
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function getFiles($id, $request=array()){
        return file::getByProperty('parentid',$id,$request);
	}

    /**
     * Gets the notes associated with the record.
     * @param int $id
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function getNotes($id, $request=array()){
        $request['pagenumber']='1';
        $request['pagesize']='100000';
        $request['sortby']='entered';
        return note::getByProperty('parentid',$id,$request);
    }

    /**
     * Gets the audio recordings associated with the record.
     * @param int $id
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function getRecordings($id, $request=array()){
        $request['pagenumber']='1';
        $request['pagesize']='100000';
        $request['sortby']='id';
        $request['sortdescending']='yes';
        return audiorecording::getByProperty('parentid',$id,$request);
    }

    /**
     * Forces the files, notes, and audio recordings attached to an object to have the same project ID
     * as that object.
     * @param $parentId
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function updateFileNoteAndRecordingProjectIds($parentId){
        file::setProjectIdsToMatchParent($parentId);
        note::setProjectIdsToMatchParent($parentId);
        audiorecording::setProjectIdsToMatchParent($parentId);
    }

    /**
     * Creates a new record
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function create($request=array()){
	    $className=get_called_class();
	    return baseobject::doCreate($request, $className);
	}

    /**
     * @param array $request
     * @param $className
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function createByClassName($request,$className){
	    return static::doCreate($request, $className);
	}

    /**
     * @param $request
     * @param $className
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
	private static function doCreate($request, $className){
		if(null==$className){ $className=get_called_class(); }
		$fields=array();
		if(!isset($request['projectid'])){
			throw new BadRequestException('Project ID not provided.');
		}
		$fieldValidations=forward_static_call($className.'::getFieldValidations');
		foreach($fieldValidations as $fname=>$validations){
			if(isset($request[$fname])){
				validator::validate($fname, $request[$fname], $validations);
				$fields[]=$fname;
			} else {
				validator::validate($fname, null, $validations);
			}
		}
		$keys='`'.implode('`,`', $fields).'`';
		$values=':'.implode(',:',$fields);
		static::beforeCreateHook($request, $className);
		$baseObjectId=self::createBaseObject($request['projectid'], $className);
        /** @noinspection SqlInsertValues */
        /** @noinspection SqlResolve */
        $sqlStatement='INSERT INTO '.$className.'(id,projectid,'.$keys.') VALUES(:id,:projectid,'.$values.')';
		$params=array(':id'=>$baseObjectId, ':projectid'=>$request['projectid']);
		foreach($fields as $f){
			if(isset($request[$f])){
				$params[':'.$f]=$request[$f];
			}
		}
		database::query($sqlStatement,$params);
		if('file'==$className){
            $created=array('id'=>$baseObjectId, 'projectid'=>$request['projectid']);
		    $ret=array('type'=>$className, 'created'=>$created);
		} else {
		    $created=forward_static_call(array($className,'getById'), $baseObjectId);
            $ret=array('type'=>$className, 'created'=>$created);
        }
        static::afterCreateHook($created, $className);
		return $ret;
	}


    /**
     * Updates one or more fields for a given record. Fields are updated if the field name is found in the fields array of the calling class.
     * @param int id The ID of the record
     * @param array $request
     * @return array The updated record
     * @throws BadRequestException if one or more fields is invalid
     * @throws ForbiddenException if user can read but not update the record
     * @throws ServerException
     * @throws NotFoundException
     */
	public static function update($id,$request=array()){
		$className=get_called_class();
		if(isset($request['_classname'])){
		    $className=$request['_classname'];
		    unset($request['_classname']);
		}

		if(!preg_match('/^[A-Za-z0-9_-]*$/',$className)){
			throw new BadRequestException('Bad class name got to baseobject::update');
		}
		if(!session::isAdmin() && !$className::canUpdate($id)){
		    $forceSharedProject=(isset($className::$sharedProjectOverride) && $className::$sharedProjectOverride);
            /** @noinspection SqlResolve */
            $sqlStatement='SELECT projectid FROM '.$className.',project 
					WHERE '.$className.'.projectid=project.id 
						AND '.$className.'.id=:id '.static::getProjectClause('update',$forceSharedProject);
			$params=array(':id'=>$id);
			$result=database::queryGetOne($sqlStatement,$params);
			if(empty($result)){
				//id doesn't exist, or no update permissions
				self::getbyid($id); //will throw NotFoundException if not found or no read permissions
				//still here? No exception was thrown above. Object exists, we can read it, we can't update it.
				throw new ForbiddenException("You do not have permission to update that record");	
			}
		}

        //now we can get on with it
        $fieldValidations=forward_static_call_array(array($className, 'getFieldValidations'), []);
		foreach($fieldValidations as $field=>$validations){
			if(!preg_match('/^[A-Za-z0-9_-]*$/',$field)){
				throw new BadRequestException('Bad field name got to baseobject::update');
			}
		}
		static::beforeUpdateHook($id, $request, $className);
        foreach($fieldValidations as $field=>$validations){
			if(isset($request[$field])){
				validator::validate($field,$request[$field],$validations); //will throw exception if invalid
				$params=array(
					':id'=>$id,
					':newvalue'=>$request[$field]
				);			
				database::query('UPDATE `'.$className.'` SET `'.$field.'`=:newvalue WHERE id=:id', $params);
			}
		}
		//Set last edit user and time
		$now=new DateTime();
		$now->setTimezone(new DateTimeZone('GMT'));
		$params=array(
			':id'=>$id,
			':lasteditor'=>session::getUserId(),
			':lastedittime'=>$now->format('Y-m-d H:i:s')
		);
		database::query('UPDATE baseobject SET lasteditor=:lasteditor, lastedittime=:lastedittime WHERE id=:id', $params);
		static::afterUpdateHook($id, $className);
		return array('updated'=>static::getById($id));
	}


    /**
     * Deletes the record with the specified ID, or throws an exception. Note that this does an SQL DELETE of the row in the
     * baseobject table ONLY - it relies on the correct ON DELETE CASCADE behaviour being set in the child table definitions.
     * @param int id The database ID of the object to delete
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException if the user can read but not delete the record
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function delete($id){
	    $className=get_called_class();
        /** @noinspection PhpUndefinedMethodInspection */
        if(!session::isAdmin() && !$className::canDelete($id)){
	        $forceSharedProject=(isset($className::$sharedProjectOverride) && $className::$sharedProjectOverride);
            /** @noinspection SqlResolve */
            $sqlStatement='SELECT projectid FROM '.$className.',project
					WHERE '.$className.'.projectid=project.id
						AND '.$className.'.id=:id '.static::getProjectClause('delete',$forceSharedProject);
	        $params=array(':id'=>$id);
	        $result=database::queryGetOne($sqlStatement,$params);
	        if(empty($result)){
	            //id doesn't exist, or no delete permissions
	            self::getbyid($id); //will throw NotFoundException if not found or no read permissions
	            //still here? No exception was thrown above. Object exists, we can read it, we can't delete it.
	            throw new ForbiddenException("You do not have permission to delete that record");
	        }
	    }
        $object=static::getById($id);
        static::beforeDeleteHook($object, $className);
		$sqlStatement='DELETE FROM baseobject WHERE id=:id';
		$params=array(':id'=>$id);
		database::query($sqlStatement,$params);
		//if affected rows, OK, else
		//get by id
        static::afterDeleteHook($object, $className);
		return array('deleted'=>$id);
	}

    /**
     * Whether the current user can create records in any project.
     * This default implementation returns true if the user (a) is an administrator, or (b) has create permissions on any project.
     * @throws BadRequestException
     */
	public static function canCreate(){
		return session::isAdmin() || !empty(session::getCreateProjects());
	}

    /**
     * Whether the current user can update the record with this ID.
     * This default implementation returns true if the user (a) is an administrator, or (b) has update permissions on the record's project.
     * @param $id
     * @return bool
     * @throws BadRequestException
     * @throws ServerException
     * @throws NotFoundException
     */
	public static function canUpdate($id){
	    if(session::isAdmin()){ return true; }
		$projects=session::getUpdateProjects();
		$item=static::getById($id);
		return in_array($item['projectid'], $projects);
	}


    /**
     * Whether the current user can delete the record with this ID.
     * This default implementation returns true if the user (a) is an administrator, or (b) has delete permissions on the record's project.
     * @param $id
     * @return bool
     * @throws BadRequestException
     * @throws ServerException
     * @throws NotFoundException
     */
	public static function canDelete($id){
		if(session::isAdmin()){ return true; }
		if(!config::get('allow_delete')){ return false; }
		$projects=session::getDeleteProjects();
		$item=static::getById($id);
		return in_array($item['projectid'], $projects);
	}

    /**
     * Calls a class method with the supplied arguments, throwing a BadRequestException if class or method does not exist.
     * @param string $className The class name
     * @param string $methodName The method name
     * @param array $args The arguments to pass to the method
     * @return mixed The results of the method call
     * @noinspection PhpUnused*@throws BadRequestException if the class or method does not exist
     * @throws BadRequestException
     */
	public static function callOrThrow($className, $methodName, $args){
	    if(!class_exists($className)){
	        throw new BadRequestException('Class '.$className.' does not exist');
	    }
	    if(!method_exists($className, $methodName)){
	        throw new BadRequestException('Class '.$className.': Method '.$methodName.' does not exist');
	    }
	    return forward_static_call_array(array($className, $methodName), $args);
	}
	
	public static function canMakeQrCodes(){
	    if(!extension_loaded('gd')){ return false; }
	    if(PHP_MAJOR_VERSION<7){ return false; }
	    if(PHP_MAJOR_VERSION==7 && PHP_MINOR_VERSION<2){ return false; }
	    return true;
	}

    /**
     * Creates a QR code to the specified URL, and attaches it to the object as a file.
     * @param $id int The ID of the object.
     * @param $url string The URL to link to.
     * @param $filename string The filename to use for the QR code.
     * @param $description string The description of the code, to show in the parent object's Files tab
     * @return array|bool
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function makeQrCode($id, $url, $filename, $description){
	    if(!static::canMakeQrCodes()){ return false; }
	    $saveTo=rtrim(config::get('core_filestore'),'/').'/qr_'.str_replace('.','_',$filename).'_'.str_replace(' ','_',microtime()).'.png';

	    $generator=new barcode_generator();
	    $image=$generator->render_image('qr-m', $url, array());
	    imagepng($image, $saveTo);
	    imagedestroy($image);
	    $generator=null;
	    
 	    session::becomeAdmin();
	    $file=file::createFromTempFile(array(
	        'parentid'=>$id,
	        'mimetype'=>'image/png',
	        'filename'=>$filename,
	        'description'=>$description,
	        'tempfile'=>$saveTo
	    ));
	    session::revertAdmin();
	    return $file;
	}

    //Hook functions

    /**
     * @param $request
     * @throws ServerException
     */
    public static function beforeCreateHook($request, $className=null){
        static::doHook('before','create',array(
            'request'=>$request
        ), $className);
    }

    /**
     * @param $createdObject
     * @throws ServerException
     */
    public static function afterCreateHook($createdObject, $className=null){
        static::doHook('after','create',array(
            'created'=>$createdObject,
            'id'=>$createdObject['id']
        ), $className);
    }

    /**
     * @param $id
     * @param $request
     * @throws ServerException
     */
    public static function beforeUpdateHook($id, $request, $className=null){
        static::doHook('before','update',array(
            'id'=>$id,
            'request'=>$request
        ), $className);
    }

    /**
     * @param $id
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function afterUpdateHook($id, $className=null){
        $updatedObject=static::getById($id);
        static::doHook('after','update',array(
            'updated'=>$updatedObject,
            'id'=>$id
        ), $className);
    }

    /**
     * @param $objectToDelete
     * @throws ServerException
     */
    public static function beforeDeleteHook($objectToDelete, $className=null){
        static::doHook('before','delete',array(
            'toDelete'=>$objectToDelete,
            'id'=>$objectToDelete['id']
        ), $className);
    }

    /**
     * @param $deletedObject
     * @throws ServerException
     */
    public static function afterDeleteHook($deletedObject, $className=null){
        static::doHook('after','delete',array(
            'deleted'=>$deletedObject,
            'id'=>$deletedObject['id']
        ), $className);
    }

    /**
     * Performs actions before calling the IceBear API method.
     * @param $when string 'before' or 'after'
     * @param $action string What is being done (create,update,delete)
     * @param $args array The arguments to be made available in the hook file
     * @param null|string $className
     * @throws ServerException
     */
    private static function doHook($when, $action, $args, $className=null){
        if(null===$className){ $className=get_called_class(); }
        if(!in_array($when,array('before','after'))){
            throw new ServerException('Bad when in doHook: '.$when.' (should be one of before,after)');
        }
        $hookActions=array('create','update','delete');
        if(!in_array($action, $hookActions)){
            throw new ServerException('Bad action in doHook: '.$action.' (should be one of '.implode(',',$hookActions).')');
        }
        $filename=$className.'_'.$when.'_'.$action;
        if(!preg_match('/^[a-z]+_[a-z]+_[a-z]+$/', $filename)){
            throw new ServerException('Bad hook filename in doHook: '.$filename);
        }
        $hookFile=rtrim($_SERVER['DOCUMENT_ROOT'],'/').'/api/_hooks/'.$filename;
        if(file_exists($hookFile.'.php')){
            /** @noinspection PhpIncludeInspection */
            include_once($hookFile.'.php');
        }
        $count=1;
        while(file_exists($hookFile.'_'.$count.'.php')){
            /** @noinspection PhpIncludeInspection */
            include_once($hookFile.'_'.$count.'.php');
            $count++;
        }
    }

}
